import os
import re
from datetime import timedelta
from collections import namedtuple

import click

_DomainWebrootTuple = namedtuple('DomainWebrootTuple', ['domain', 'webroot'])
_WebrootTuple = namedtuple('WebrootTuple', ['remote_user', 'remote_host', 'remote_port', 'path'])
_WebrootTuple.__new__.__defaults__ = (None,) * len(_WebrootTuple._fields)


class _DomainWebrootType(click.ParamType):
    _re = re.compile((r'^(?P<domain>[^:]+)'
                      r'(?::'
                      r'(?:'
                      r'(?:(?P<remote_user>[^@]+)@)?'
                      r'(?P<remote_host>[^@:]+)'
                      r'(?::(?P<remote_port>\d+))?:'
                      r')?'
                      r'(?P<path>[^@:]+)'
                      r')?$'))

    domain = None
    webroot = None

    def convert(self, value, param, ctx):
        if isinstance(value, _DomainWebrootTuple):
            return value

        match = self._re.match(value)
        if not match:
            self.fail('could not parse %s as %s' % (value, self.get_metavar(None)))

        remote_port = match.group('remote_port')
        if remote_port:  # will never be empty according to regex
            remote_port = int(remote_port)

        if match.group('path'):
            path = match.group('path')
            if not match.group('remote_host'):
                path = WritablePathType(os.path.expanduser(path))
            webroot = _WebrootTuple(
                remote_user=match.group('remote_user'),
                remote_host=match.group('remote_host'),
                remote_port=remote_port,
                path=path)
        else:
            webroot = None
        return _DomainWebrootTuple(domain=match.group('domain'), webroot=webroot)

    def get_metavar(self, param):
        return 'DOMAIN[[[USER@]HOST[:PORT]]:WEBROOT]'


class _TimespanType(click.ParamType):
    _re = re.compile(r'^(?P<amount>\d+)(?P<unit>h|d|w)$')
    _unitmap = {
        'h': 'hours',
        'd': 'days',
        'w': 'weeks',
    }

    def convert(self, value, param, ctr):
        if isinstance(value, timedelta):
            return value
        match = self._re.match(value)
        if not match:
            self.fail('could not parse %s as timespan' % value)
        return timedelta(**{self._unitmap[match.group('unit')]: int(match.group('amount'))})

    def get_metavar(self, param):
        return 'TIME'


DomainWebrootType = _DomainWebrootType()
TimespanType = _TimespanType()
WritablePathType = click.Path(exists=True, writable=True, readable=False, file_okay=False, dir_okay=True,
                              resolve_path=True)



import os
from datetime import datetime, timedelta

import pytest
from mock import Mock, ANY, call

import wile
from wile import lazyclient
from wile import argtypes
from wile import cert

FIXTURE_DATA_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'certs',
)


def test_generate_domain_and_webroot_lists_from_args__fail(inside_tmpdir):
    ctx_mock = Mock(spec_set=['exit'])
    (dl, wrl) = cert._generate_domain_and_webroot_lists_from_args(ctx_mock, [
                    argtypes.DomainWebrootType('example.org'),
                ])
    ctx_mock.exit.assert_called_with(1)


@pytest.mark.parametrize("args, expected_dl, expected_wrl", [
    (['example.org:folder1'], ['example.org'], ['folder1']),
    (['example.org:folder1', 'foo.example.org'], ['example.org', 'foo.example.org'], ['folder1', 'folder1']),
    (['example.org:folder1', 'foo.example.org:folder2'], ['example.org', 'foo.example.org'], ['folder1', 'folder2']),
])
def test_generate_domain_and_webroot_lists_from_args__success(args, expected_dl, expected_wrl, inside_tmpdir):
    ctx_mock = Mock(spec_set=['exit'])

    os.mkdir('folder1')
    os.mkdir('folder2')

    (dl, wrl) = cert._generate_domain_and_webroot_lists_from_args(ctx_mock, list(
                    map(argtypes.DomainWebrootType, args),
                ))
    ctx_mock.exit.assert_not_called()

    assert dl == expected_dl
    assert wrl == list(map(lambda f: argtypes._WebrootTuple(path=os.path.join(os.path.abspath('.'), f)), expected_wrl))


@pytest.mark.parametrize("args", [
    ['cert'],
    ['cert', '--help'],
    ['cert', 'revoke', '--help'],
])
def test_wile_cert__noargs_and_help(args, clirunner, logcapture, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    result = clirunner.invoke(wile.wile, args=args, obj=lazyclient.LazyClient())
    assert result.exit_code == 0
    assert result.output.startswith('Usage:')
    logcapture.check()
    assert os.listdir(os.path.expanduser('~')) == []
    assert os.listdir(os.curdir) == []
    assert not acmeclientmock.called


def test_wile_cert_revoke__noargs(clirunner, logcapture, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    result = clirunner.invoke(wile.wile, args=['cert', 'revoke'], obj=lazyclient.LazyClient())
    assert result.exit_code == 2
    assert result.output.startswith('Usage:')
    logcapture.check()
    assert os.listdir(os.path.expanduser('~')) == []
    assert os.listdir(os.curdir) == []
    assert not acmeclientmock.called


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DATA_DIR, 'cert1.crt'),
    os.path.join(FIXTURE_DATA_DIR, 'cert2.crt'),
)
def test_wile_cert_revoke(datafiles, clirunner, logcapture, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    certs = list(map(str, datafiles.listdir()))
    result = clirunner.invoke(wile.wile, args=['cert', 'revoke']+certs, obj=lazyclient.LazyClient())
    assert result.exit_code == 0
    logcapture.check(
        ('wile', 'WARNING',
         'no account key found; creating a new 2048 bit key in %s' % os.path.expanduser('~/.wile/account.key'))
    )
    assert os.listdir(os.path.expanduser('~')) == ['.wile']
    assert os.listdir(os.path.expanduser('~/.wile')) == ['account.key']
    assert os.listdir(os.curdir) == []
    acmeclientmock.assert_called_once_with(wile.LETSENCRYPT_URL, ANY)
    acmeclientmock.return_value.revoke.assert_has_calls([call(ANY, 0), call(ANY, 0)])


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DATA_DIR, 'cert_example.com.crt')
)
def test_is_valid_and_unchanged__expired(datafiles, fixed_datetime_monkeypatch):
    # time after cert validity
    fixed_datetime_monkeypatch(datetime(2017, 10, 17, 0, 0, 0))
    res = cert._is_valid_and_unchanged(str(datafiles / 'cert_example.com.crt'),
                                       ['example.com', 'www.example.com', 'test.example.com'],
                                       timedelta(0))
    assert res is False


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DATA_DIR, 'cert_example.com.crt')
)
def test_is_valid_and_unchanged__new_domain(datafiles, fixed_datetime_monkeypatch):
    # time after cert validity
    fixed_datetime_monkeypatch(datetime(2017, 10, 16, 0, 0, 0))
    res = cert._is_valid_and_unchanged(str(datafiles / 'cert_example.com.crt'),
                                       ['example.com', 'www.example.com', 'test.example.com', 'new.example.com'],
                                       timedelta(0))
    assert res is False


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DATA_DIR, 'cert_example.com.crt')
)
def test_is_valid_and_unchanged__removed_domain(datafiles, fixed_datetime_monkeypatch):
    # time after cert validity
    fixed_datetime_monkeypatch(datetime(2017, 10, 16, 0, 0, 0))
    res = cert._is_valid_and_unchanged(str(datafiles / 'cert_example.com.crt'),
                                       ['example.com', 'www.example.com'],
                                       timedelta(0))
    assert res is False


@pytest.mark.datafiles(
    os.path.join(FIXTURE_DATA_DIR, 'cert_example.com.crt')
)
def test_is_valid_and_unchanged(datafiles, fixed_datetime_monkeypatch):
    # time after cert validity
    fixed_datetime_monkeypatch(datetime(2017, 10, 16, 0, 0, 0))
    res = cert._is_valid_and_unchanged(str(datafiles / 'cert_example.com.crt'),
                                       ['example.com', 'www.example.com', 'test.example.com'],
                                       timedelta(0))
    assert res is True


def test_store_webroot_validation__local(inside_tmpdir):
    webroot = Mock(spec_set=['path', 'remote_host'], **{
        'path': inside_tmpdir[0].name,
        'remote_host': None,
    })

    challb = Mock(spec_set=['path', 'URI_ROOT_PATH'], **{
        'path': '.well-known/acme-challenge/sometoken',
        'URI_ROOT_PATH': '.well-known/acme-challenge',
    })
    cert._store_webroot_validation(webroot, None, challb, 'some validation content')

    assert os.listdir(os.curdir) == ['.well-known']
    assert os.listdir('.well-known') == ['acme-challenge']
    assert os.listdir('.well-known/acme-challenge') == ['sometoken']
    assert open('.well-known/acme-challenge/sometoken').read() == 'some validation content'



import os
import sys
import logging
import datetime
from functools import partial

from mock import Mock, PropertyMock
import pytest
from click.testing import CliRunner
from testfixtures import LogCapture
from acme import client

if sys.version_info.major < 3:
    import backports.tempfile as tempfile
else:
    import tempfile


@pytest.fixture(scope="function")
def clirunner(inside_tmpdir):
    return CliRunner()


@pytest.fixture()
def logcapture():
    with LogCapture(level=logging.WARN) as lc:
        yield lc


@pytest.fixture(scope="function")
def inside_tmpdir(monkeypatch):
    tmpdir = tempfile.TemporaryDirectory()
    homedir = tempfile.TemporaryDirectory()

    monkeypatch.setenv('HOME', homedir.name)
    os.chdir(tmpdir.name)

    return (tmpdir, homedir)


@pytest.fixture()
def fixed_datetime_monkeypatch(monkeypatch):
    def _factory(now):
        class _fixed_datetime(datetime.datetime):
            @classmethod
            def now(cls):
                return now
        monkeypatch.setattr(datetime, 'datetime', _fixed_datetime)
    return _factory


def _acmeclientcmock(monkeypatch, ask_for_tos=True, emails='', phones=''):
    '''
    This mocks most of the acme client's functionality that we use.
    '''

    if ask_for_tos:
        TOS1 = 'http://some.tos.url'
        TOS2 = 'http://some.other.tos.url'
    else:
        TOS1 = TOS2 = 'http://some.tos.url'

    original_contacts = dict(
        emails=(emails,),
        phones=(phones,),
    )

    def upd_reg(_, new_reg):
        original_contacts['emails'] = new_reg.emails
        original_contacts['phones'] = new_reg.phones

    acmeRegistrationMock = Mock(spec_set=['body', 'terms_of_service'], **{
        'body': Mock(spec_set=['agreement', 'emails', 'phones'], **{
            'agreement': TOS1,
        }),
        'terms_of_service': TOS2,
    })
    type(acmeRegistrationMock.body).emails = PropertyMock(side_effect=lambda: original_contacts['emails'])
    type(acmeRegistrationMock.body).phones = PropertyMock(side_effect=lambda: original_contacts['phones'])

    acmeClientInstanceMock = Mock(spec_set=['register', 'agree_to_tos', 'update_registration',
                                            'query_registration', 'revoke'], **{
        'register.return_value': acmeRegistrationMock,
        'update_registration.return_value': acmeRegistrationMock,
        'update_registration.side_effect': upd_reg,
        'query_registration.return_value': acmeRegistrationMock,
        'agree_to_tos.return_value': None,
    })
    acmeClientClassMock = Mock(spec_set=[], return_value=acmeClientInstanceMock)

    monkeypatch.setattr(client, 'Client', acmeClientClassMock)

    return acmeClientClassMock


@pytest.fixture(scope="function")
def acmeclientmock_factory(monkeypatch):
    return partial(_acmeclientcmock, monkeypatch)



import os
import types
import logging

import paramiko

logger = logging.getLogger(__name__)

_cache = dict()


def cachedSFTPfactory(host, port, user, private_key):
    key = (host, port, user)

    if key in _cache:
        return _cache[key]
    else:
        ssh = paramiko.SSHClient()
        ssh.load_host_keys(os.path.expanduser('~/.ssh/known_hosts'))
        ssh.connect(hostname=host, port=port or 22,
                    username=user, key_filename=private_key,
                    password=os.getenv('WILE_SSH_PASS'))
        sftp = ssh.open_sftp()
        sftp._ssh = ssh  # workaround for paramiko issue #344
        sftp.makedirs = types.MethodType(_makedirs, sftp)
        _cache[key] = sftp

        return sftp


def _makedirs(self, path):
    head, tail = os.path.split(path)
    if head and head != '/':
        self.makedirs(head)
    try:
        logger.debug('creating directory %s on remote host', path)
        self.mkdir(path)
    except OSError:
        # sftp doesn't return any useful information, so we assume it's an EEXIST and hope for the best ;)
        pass



import os

import pytest
import click

import wile


def test_wile__get_or_gen_key(inside_tmpdir, logcapture, monkeypatch):
    account_key_path = 'account.key'
    account_key_size = 2048

    monkeypatch.setattr(click, 'prompt', lambda *args, **kwargs: u'somepassword')

    assert os.listdir(os.curdir) == []
    key1 = wile.get_or_gen_key(None, account_key_path, account_key_size)
    logcapture.check(
        ('wile', 'WARNING', 'no account key found; creating a new 2048 bit key in account.key')
    )
    logcapture.clear()
    assert os.listdir(os.curdir) == [account_key_path]

    key2 = wile.get_or_gen_key(None, account_key_path, account_key_size)
    logcapture.check()
    assert key1 == key2


@pytest.mark.parametrize("args", [
    [],
    ['--help'],
])
def test_wile__no_args_or_help(args, clirunner, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    result = clirunner.invoke(wile.wile, args=args)
    assert result.output_bytes.startswith(b'Usage:')
    assert result.exit_code == 0
    assert os.listdir(os.path.expanduser('~')) == []
    assert os.listdir(os.curdir) == []  # ensure it's a noop
    assert not acmeclientmock.called


def test_wile__version(clirunner, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    result = clirunner.invoke(wile.wile, args=['--version'])
    assert ('version %s' % wile._version) in str(result.output_bytes)
    assert result.exit_code == 0
    assert os.listdir(os.path.expanduser('~')) == []
    assert os.listdir(os.curdir) == []  # ensure it's a noop
    assert not acmeclientmock.called



import os
import logging
from functools import partial

import setuptools_scm
import click
import josepy as jose
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa

from . import cert
from . import reg
from .lazyclient import LazyClient

logger = logging.getLogger(__name__)

try:  # if we're in the checked out tree, use setuptools_scm
    _version = setuptools_scm.get_version()
except LookupError:  # otherwise click will attempt to find it via pkg_resources
    _version = None

LETSENCRYPT_URL = 'https://acme-v01.api.letsencrypt.org/directory'
LETSENCRYPT_STAGING_URL = 'https://acme-staging.api.letsencrypt.org/directory'


@click.group()
@click.pass_context
@click.version_option(version=_version)
@click.option('--staging', is_flag=True, default=False, help='use letsencrypt\'s staging server (for testing)')
@click.option('--directory-url', metavar='URL', default=LETSENCRYPT_URL,
              show_default=True, help='URL for alternative ACME directory (will be overriden by --staging)')
@click.option('--account-key', 'account_key_path', type=click.Path(dir_okay=False, allow_dash=True),
              default='~/.wile/account.key', show_default=True, help='path to existing account key')
@click.option('--new-account-key-size', type=int, metavar='BITS', default=2048, show_default=True,
              help='bit size to use when creating a new account key; ignored for existing keys')
@click.option('--verbose', '-v', count=True, help='be more verbose; can be passed multiple times')
def wile(ctx, directory_url, staging, account_key_path, new_account_key_size, verbose):
    '''
    Simple client for ACME servers (e.g. letsencrypt).
    '''

    if verbose > 1:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose > 0:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    if staging:
        directory_url = LETSENCRYPT_STAGING_URL

    account_key_callback = partial(get_or_gen_key, ctx, account_key_path, new_account_key_size)

    logger.debug('connecting to ACME directory at %s', directory_url)
    ctx.obj.init(directory_url, account_key_callback)


wile.add_command(cert.cert)
wile.add_command(reg.register)


def get_or_gen_key(ctx, account_key_path, new_account_key_size):
    account_key_path = os.path.expanduser(account_key_path)
    if os.path.exists(account_key_path):
        logger.debug('opening existing account key %s', account_key_path)
        with open(account_key_path, 'rb') as key_file:
            key_contents = key_file.read()
            try:
                try:
                    account_key = jose.JWKRSA(key=serialization.load_pem_private_key(key_contents, None,
                                              default_backend()))
                except TypeError:  # password required
                    password = click.prompt('Password for %s' % account_key_path, hide_input=True, default=None)
                    key = serialization.load_pem_private_key(key_contents, password.encode('utf-8'), default_backend())
                    account_key = jose.JWKRSA(key=key)
            except ValueError as e:
                logger.error('could not open key %s: %s', account_key_path, e)
                ctx.exit(1)
    else:
        logger.warn('no account key found; creating a new %d bit key in %s', new_account_key_size, account_key_path)
        account_key = jose.JWKRSA(key=rsa.generate_private_key(
            public_exponent=65537,
            key_size=new_account_key_size,
            backend=default_backend()))
        try:
            os.makedirs(os.path.dirname(account_key_path), 0o750)
        except os.error:
            pass  # dir already exists

        encryption_algorithm = ask_for_password_or_no_crypto(account_key_path)
        with open(account_key_path, 'wb') as key_file:
            key_file.write(account_key.key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=encryption_algorithm
            ))
    return account_key


def ask_for_password_or_no_crypto(key_path):
    # we can't use prompt's "default" and "value_proc" arguments because we monkeypatch prompt in test_wile.py
    password = click.prompt('(optional) Password for %s' % key_path, default='',
                            hide_input=True, confirmation_prompt=True, show_default=False)
    if password:
        return serialization.BestAvailableEncryption(password.encode('utf-8'))
    else:
        return serialization.NoEncryption()


def main():
    return wile(obj=LazyClient())



import atexit
import os
import logging
import errno
import datetime
from collections import OrderedDict
from six import b
from six.moves import range

import click
from OpenSSL import crypto

from acme import challenges
from acme import messages
from acme import errors
from josepy.util import ComparableX509

from . import reg
from . import argtypes
from . import sftp_helper

logger = logging.getLogger(__name__)

# Taken from https://tools.ietf.org/html/rfc5280#section-5.3.1
# Not all all supported by letsencrypt's boulder.
REVOCATION_REASONS = OrderedDict((
    ('unspecified', 0),
    ('keyCompromise', 1),
    ('CACompromise', 2),
    ('affiliationChanged', 3),
    ('superseded', 4),
    ('cessationOfOperation', 5),
    ('certificateHold', 6),
    ('removeFromCRL', 8),
    ('privilegeWithdrawn', 9),
    ('AACompromise', 10),
))


@click.group()
def cert():
    '''
    Commands for certificate management.
    '''


@cert.command()
@click.pass_context
@click.option('--with-chain/--separate-chain', is_flag=True, default=True, show_default=False,
              help='Whether to include the certificate\'s chain in the output certificate; --separate-chain implies a '
                   'separate .chain.crt file, containing only the signing certificates up to the root '
                   '[default: with chain]')
@click.option('--key-size', '-s', metavar='SIZE', type=int, default=2048, show_default=True,
              help='Size in bits for the generated certificate\'s key')
@click.option('--output-dir', metavar='DIR', type=argtypes.WritablePathType, default='.',
              help='Where to store created certificates (default: current directory)')
@click.option('--basename', metavar='BASENAME',
              help='Basename to use when storing output: BASENAME.crt and BASENAME.key [default: first domain]')
@click.option('--key-digest', metavar='DIGEST', default='sha256', show_default=True,
              help='The digest to use when signing the request with its key (must be supported by openssl)')
@click.option('--min-valid-time', type=argtypes.TimespanType, metavar='TIMESPAN', default='1w', show_default=True,
              help='If a certificate is found and its expiration lies inside of this timespan, it will be '
                   'automatically requested and overwritten; otherwise no request will be made. The format for this '
                   'option is "[DIGITS][UNIT]", where UNIT is one of "h" for hours, "d" for days and "w" for weeks.')
@click.option('--force', is_flag=True, default=False, show_default=True,
              help='Whether to force a request to be made, even if a valid certificate is found')
@click.option('--ssh-private-key', type=click.Path(file_okay=True, dir_okay=False), default=None, show_default=True,
              help='Path to SSH private key when using remote webroots. If not provided, the default paths will be '
                   'searched. If an SSH agent is running, it will also be queried independently of this setting.')
@click.option('--symlink-domains', is_flag=True, default=False, show_default=True,
              help='Create symlinks for each additional domain, pointing to the created certificate and key pair. If '
                   'the --basename option is provided, an additional symlink for the first domain will also be '
                   'created.')
@click.option('--reuse-key', is_flag=True, default=False, show_default=True,
              help='By default, wile creates a new key for each new certificate. This options enables key reuse for '
                   'multiple certificate requests. If no existing key is found, a new one will be generated and the '
                   'option will be ignored. The --basename and --output-dir options are used when trying to find an '
                   'existing key')
@click.argument('domainroots', 'DOMAIN[:WEBROOT]', type=argtypes.DomainWebrootType, metavar='DOMAIN[:WEBROOT]',
                nargs=-1, required=True)
def request(ctx, domainroots, with_chain, key_size, output_dir, basename, key_digest,
            min_valid_time, force, ssh_private_key, symlink_domains, reuse_key):
    '''
    Request a new certificate for the provided domains and respective webroot paths.
    If a webroot is not provided for a domain, the one supplied for the previous domain is used.\n
    Each WEBROOT may be a local path or a remote path in the following format:
    [[[USER@]HOST[:PORT]:]PATH].\n
    In the second case an SSH connection will be made to the remote HOST and the ACME challanges will
    be remotely verified.
    '''

    regr = ctx.invoke(reg.register, quiet=True, auto_accept_tos=True)

    domain_list, webroot_list = _generate_domain_and_webroot_lists_from_args(ctx, domainroots)
    alt_domain_list = basename and domain_list[:] or domain_list[1:]
    basename = basename or domain_list[0]
    keyfile_path = os.path.join(output_dir, '%s.key' % basename)
    certfile_path = os.path.join(output_dir, '%s.crt' % basename)
    chainfile_path = os.path.join(output_dir, '%s.chain.crt' % basename)

    if os.path.exists(certfile_path):
        if not force and _is_valid_and_unchanged(certfile_path, domain_list, min_valid_time):
            logger.info('found existing valid certificate (%s); not requesting a new one', certfile_path)
            ctx.exit(0)
        elif force:
            logger.info('found existing valid certificate (%s), but forcing renewal on request', certfile_path)
        else:
            logger.info('existing certificate (%s) will expire inside of renewal time (%s) or has changes; '
                        'requesting new one', certfile_path, min_valid_time)
            force = True

    authzrs = _generate_validation_requests(domain_list, webroot_list, ctx, regr, ssh_private_key)

    key_reused = False
    if reuse_key and os.path.isfile(keyfile_path):
        with open(keyfile_path, 'r') as f:
            key = crypto.load_privatekey(crypto.FILETYPE_PEM, f.read())
        key_reused = True
    else:
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, key_size)

    csr = _generate_csr(key, key_digest, domain_list)

    crt = _poll_validations_and_fetch_crt(ctx, csr, authzrs)

    # write optional chain
    chain = ctx.obj.acme.fetch_chain(crt)
    certs = [crt.body]
    if with_chain:
        certs.extend(chain)
    else:
        if not force and os.path.exists(chainfile_path):
            _confirm_overwrite(chainfile_path)

        with open(chainfile_path, 'wb') as chainfile:
            for crt in chain:
                chainfile.write(crypto.dump_certificate(crypto.FILETYPE_PEM, crt))

    # write cert
    with open(certfile_path, 'wb') as certfile:
        for crt in certs:
            certfile.write(crypto.dump_certificate(crypto.FILETYPE_PEM, crt))

    # write key
    if not key_reused and not force and os.path.exists(keyfile_path):
        _confirm_overwrite(keyfile_path)

    with open(keyfile_path, 'wb') as keyfile:
        os.chmod(keyfile.name, 0o640)
        keyfile.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))

    if symlink_domains:
        for domain in alt_domain_list:
            certfile_link = os.path.join(output_dir, '%s.crt' % domain)
            keyfile_link = os.path.join(output_dir, '%s.key' % domain)
            if not os.path.exists(certfile_link):
                logger.info('creating symlink %s -> %s', certfile_link, certfile_path)
                os.symlink(certfile_path, certfile_link)
            else:
                logger.warn('not overwriting %s with symlink', certfile_link)

            if not os.path.exists(keyfile_link):
                logger.info('creating symlink %s -> %s', keyfile_link, keyfile_path)
                os.symlink(keyfile_path, keyfile_link)
            else:
                logger.warn('not overwriting %s with symlink', keyfile_link)


def _generate_validation_requests(domain_list, webroot_list, ctx, regr, ssh_private_key):
    authzrs = list()
    for (domain, webroot) in zip(domain_list, webroot_list):
        logger.info('requesting challenge for %s in %s', domain, webroot)

        authzr = ctx.obj.acme.request_domain_challenges(domain, new_authzr_uri=regr.new_authzr_uri)
        authzrs.append(authzr)

        challb = _get_http_challenge(ctx, authzr)
        chall_response, chall_validation = challb.response_and_validation(ctx.obj.account_key)
        _store_webroot_validation(webroot, ssh_private_key, challb,
                                  chall_validation)
        ctx.obj.acme.answer_challenge(challb, chall_response)
    return authzrs


def _poll_validations_and_fetch_crt(ctx, csr, authzrs):
    try:
        logger.info('validating requests')
        crt, updated_authzrs = ctx.obj.acme.poll_and_request_issuance(csr, authzrs)
    except errors.PollError as e:
        if e.exhausted:
            logger.error('validation timed out for the following domains: %s', ', '.join(authzr.body.identifier for
                                                                                         authzr in e.exhausted))
        invalid_domains = [(e_authzr.body.identifier.value, _get_http_challenge(ctx, e_authzr).error.detail)
                           for e_authzr in e.updated.values() if e_authzr.body.status == messages.STATUS_INVALID]
        if invalid_domains:
            logger.error('validation invalid for the following domains:')
            for invalid_domain in invalid_domains:
                logger.error('%s: %s' % invalid_domain)
        ctx.exit(1)
    return crt


@cert.command()
@click.pass_context
@click.option('--reason', metavar='REASON', default='unspecified', type=click.Choice(REVOCATION_REASONS.keys()),
              show_default=True, help='reason for revoking certificate. Valid values: %s; not all are supported by '
                                      'Let\'s Encrypt' % REVOCATION_REASONS.keys())
@click.argument('cert_paths', metavar='CERT_FILE [CERT_FILE ...]', nargs=-1, required=True)
def revoke(ctx, reason, cert_paths):
    '''
    Revoke existing certificates.
    '''
    for cert_path in cert_paths:
        with open(cert_path, 'rb') as certfile:
            crt = crypto.load_certificate(crypto.FILETYPE_PEM, certfile.read())
            try:
                ctx.obj.acme.revoke(ComparableX509(crt), REVOCATION_REASONS[reason])
            except messages.Error as e:
                logger.error(e)


def _confirm_overwrite(filepath):
    click.confirm('file %s exists; overwrite?' % filepath, abort=True)


def _generate_domain_and_webroot_lists_from_args(ctx, domainroots):
    domain_list = list()
    webroot_list = list()
    webroot = None
    for domainroot in domainroots:
        if domainroot.webroot:
            webroot = domainroot.webroot
        elif webroot:
            pass  # if we already have one from the last element, just use it
        else:
            logger.error('domain without webroot: %s', domainroot.domain)
            ctx.exit(1)
        domain_list.append(domainroot.domain)
        webroot_list.append(webroot)

    return (domain_list, webroot_list)


def _get_http_challenge(ctx, authzr):
    for combis in authzr.body.combinations:
        if len(combis) == 1 and isinstance(authzr.body.challenges[combis[0]].chall, challenges.HTTP01):
            return authzr.body.challenges[combis[0]]
    ctx.fail('no acceptable challenge type found; only HTTP01 supported')


def _store_webroot_validation(webroot, ssh_private_key, challb, val):
    logger.info('storing validation of %s', webroot)

    chall_path = os.path.join(webroot.path, challb.path.strip('/'))

    if not webroot.remote_host:
        try:
            os.makedirs(os.path.join(webroot.path, challb.URI_ROOT_PATH), 0o755)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        chall_mod = os
        chall_open = open
    else:
        sftp = sftp_helper.cachedSFTPfactory(user=webroot.remote_user, host=webroot.remote_host,
                                             port=webroot.remote_port, private_key=ssh_private_key)

        sftp.makedirs(os.path.join(webroot.path, challb.URI_ROOT_PATH))

        chall_mod = sftp
        chall_open = chall_mod.open

    with chall_open(chall_path, 'wb') as outf:
        logger.info('storing validation to %s', os.path.basename(chall_path))
        outf.write(b(val))
        # TODO: this may cause a race-condition with paramiko teardown code.
        # We should probably have a more robust cleanup strategy without atexit.
        atexit.register(chall_mod.unlink, chall_path)


def _is_valid_and_unchanged(certfile_path, domains, min_valid_time):
    with open(certfile_path, 'rb') as certfile:
        crt = crypto.load_certificate(crypto.FILETYPE_PEM, certfile.read())
        # TODO: do we need to support the other possible ASN.1 date formats?
        expiration = datetime.datetime.strptime(crt.get_notAfter().decode('ascii'), r'%Y%m%d%H%M%SZ')

        # create a set of domain names in the cert (DN + SANs)
        crt_domains = {dict(crt.get_subject().get_components())[b('CN')].decode('ascii')}
        for ext_idx in range(crt.get_extension_count()):
            ext = crt.get_extension(ext_idx)
            if ext.get_short_name() == b'subjectAltName':
                # we strip 'DNS:' without checking if it's there; if it
                # isn't, the cert uses some other unsupported identifier,
                # and is definitely different from the one we're creating
                crt_domains = crt_domains.union((x.strip()[4:] for x in str(ext).split(',')))

        if datetime.datetime.now() + min_valid_time > expiration:
            logger.info('EXPIRATION')
            return False
        elif crt_domains != set(domains):
            logger.info('DOMAINS: %s != %s', crt_domains, set(domains))
            return False
        else:
            return True


def _generate_csr(key, key_digest, domains):
    csr = crypto.X509Req()
    csr.set_version(2)
    csr.set_pubkey(key)

    sans = ', '.join('DNS:{}'.format(d) for d in domains)
    exts = [crypto.X509Extension(b'subjectAltName', False, b(sans))]
    csr.add_extensions(exts)

    csr.sign(key, str(key_digest))

    return ComparableX509(csr)



from acme import client


class NotInitializedError(Exception):
    pass


class LazyClient(object):
    __client = None
    __account_key = None

    def init(self, directory_url, account_key_callback):
        self.__directory_url = directory_url
        self.__account_key_callback = account_key_callback

    @property
    def account_key(self):
        if not self.__account_key:
            self.__account_key = self.__account_key_callback()
        return self.__account_key_callback()

    @property
    def acme(self):
        if not self.__client:
            self.__client = client.Client(self.__directory_url, self.account_key)
        return self.__client



import logging

import click
from acme import messages
from acme import client
from acme import errors

logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@click.option('--email', '-e', metavar='EMAIL', help='email for contact')
# telephone contact hasn't been supported by Boulder since a while:
# https://github.com/letsencrypt/boulder/commit/03427ccb810731eecaf32f70a1d21a0001a4abe7
#
# but it's still in the specs:
# https://tools.ietf.org/html/draft-ietf-acme-acme-07#section-7.1.2
@click.option('--phone', '-p', metavar='PHONE', help='telephone number for contact; this is not supported by Let\'s'
                                       'Encrypt, but is in the ACME specs, so other providers might support it.')
@click.option('--auto-accept-tos', is_flag=True, default=False, show_default=True,
              help='Automatically accept directory\'s Terms of Service')
def register(ctx, email, phone, auto_accept_tos, quiet=False):
    '''
    Register a new account key or update an existing registration.
    '''

    logger.info('starting registration for email:"%s", phone:"%s"', email, phone)
    try:
        regr = ctx.obj.acme.register(messages.NewRegistration.from_data(email=email, phone=phone))
    except errors.ConflictError as e:
        logger.info('found existing registration for key; fetching')
        response = ctx.obj.acme.net.post(e.location, messages.UpdateRegistration())
        regr = client.Client._regr_from_response(response, uri=e.location)

    if regr.body.agreement != regr.terms_of_service:
        if not auto_accept_tos and not click.confirm('Accept Terms of Service? (%s)' % regr.terms_of_service,
                                                     default=None):
            ctx.fail('Must accept ToS to continue')
        elif auto_accept_tos:
            logger.info('automatically accepting ToS')
        ctx.obj.acme.agree_to_tos(regr)

    if (email and (email,) != regr.body.emails) or (phone and (phone,) != regr.body.phones):
        logger.info('updating contact information')
        upd_reg = messages.UpdateRegistration().from_data(email=email, phone=phone)
        try:
            regr = ctx.obj.acme.update_registration(regr, upd_reg)
        except errors.UnexpectedUpdate:
            pass  # no idea why it complains, but updates anyway
        regr = ctx.obj.acme.query_registration(regr)

    if not quiet:
        click.echo('Registration:')
        click.echo('Email: %s' % (regr.body.emails or ''))
        click.echo('Phone: %s' % (regr.body.phones or ''))
    else:
        return regr



import os

import pytest
from mock import ANY

import wile
from wile.lazyclient import LazyClient


@pytest.mark.parametrize("args, exit_code, tos_asked, url, expected_email, expected_phone", [
    (['register'], 2, True, wile.LETSENCRYPT_URL, '', ''),
    (['--staging', 'register'], 2, True, wile.LETSENCRYPT_STAGING_URL, '', ''),
    (['register', '--auto-accept-tos'], 0, False, wile.LETSENCRYPT_URL, '', ''),
    (['register', '--auto-accept-tos', '--email', 'example@example.com'], 0, False, wile.LETSENCRYPT_URL,
     'example@example.com', ''),
    (['register', '--auto-accept-tos', '--phone', '555-6666'], 0, False, wile.LETSENCRYPT_URL,
     '', '555-6666'),
])
def test_wile_register__arg_combinations(args, exit_code, tos_asked, url, expected_email, expected_phone, clirunner,
                                         logcapture, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory(tos_asked)

    result = clirunner.invoke(wile.wile, args=args, obj=LazyClient())

    assert result.exit_code == exit_code
    assert ('Accept Terms of Service?' in result.output) == tos_asked
    logcapture.check(
        ('wile', 'WARNING',
         'no account key found; creating a new 2048 bit key in %s' % os.path.expanduser('~/.wile/account.key'))
    )
    assert os.listdir(os.path.expanduser('~')) == ['.wile']
    assert os.listdir(os.path.expanduser('~/.wile')) == ['account.key']
    assert os.listdir(os.curdir) == []
    if not tos_asked:
        assert ('Email: %s' % expected_email) in result.output
        assert ('Phone: %s' % expected_phone) in result.output
    acmeclientmock.assert_called_once_with(url, ANY)


def test_wile_register__help(clirunner, logcapture, acmeclientmock_factory):
    acmeclientmock = acmeclientmock_factory()
    result = clirunner.invoke(wile.wile, args=['register', '--help'], obj=LazyClient())
    assert result.exit_code == 0
    assert result.output.startswith('Usage:')
    logcapture.check()
    assert os.listdir(os.path.expanduser('~')) == []
    assert os.listdir(os.curdir) == []
    assert not acmeclientmock.called



#!/usr/bin/python
import os
import sys
from setuptools import setup


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


config = dict(
    name="wile",
    author="Leo Antunes",
    author_email="leo@costela.net",
    description=("A stripped down Let's Encrypt (ACME) client"),
    license="GPLv3",
    keywords="letsencrypt acme ssl",
    url="https://github.com/costela/wile",
    packages=['wile'],
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
        'six',
        'acme >= 0.21.0, != 0.22.0',
        'click >= 6.0',
        'pyOpenSSL',
        'cryptography',
        'setuptools_scm',  # for run-time version-detect
        'paramiko',
        'josepy',
    ],
    tests_require=[
        'backports.tempfile;python_version<"3.0"',
        'mock',
        'pytest',
        'pytest-datafiles',
        'testfixtures',
    ],
    entry_points={
        'console_scripts': [
            'wile = wile:main',
        ],
    },
    long_description=read('README.md'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Topic :: Internet :: Name Service (DNS)',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities',
    ],
)

if 'sdist' in sys.argv or 'bdist_wheel' in sys.argv:
    config.update(dict(
        use_scm_version=True,
    ))
    config['setup_requires'] += [
        'setuptools_scm',
    ]

setup(**config)
